import SimpleITK as sitk

from stereotacticframe.frame_protocol import FrameProtocol

def apply_transform(image: sitk.Image, transform: sitk.AffineTransform, frame: FrameProtocol, interpolator=sitk.sitkLinear) -> sitk.Image:
    return sitk.Resample(image, image.GetSize(), transform, interpolator, frame.offset, image.GetSpacing(), frame.direction)
    


class ModalityNotFoundError(Exception):
    ...

class LeksellFrame:
    """Nodes and edges of Leksell fiducial box."""

    dimensions = tuple([190, 120, 120])
    offset = tuple([-10.0, 50.0, 50.0])
    direction = tuple([1, 0, 0, 0, -1, 0, 0, 0, -1])

    nodes = [
        (0, 0, 0),  # 0) right posterior cranial
        (0, -120, 0),  # 1) right anterior cranial
        (0, 0, -120),  # 2) right posterior caudal
        (0, -120, -120),  # 3) right anterior caudal
        (190, 0, 0),  # 4) left posterior cranial
        (190, -120, 0),  # 5) left anterior cranial
        (190, 0, -120),  # 6) left posterior caudal
        (190, -120, -120),  # 7) left anterior caudal
    ]

    ct_edges = [
        (0, 2),
        (0, 3),
        (1, 3),
        (4, 6),
        (4, 7),
        (5, 7),
    ]

    mr_edges = [
        (0, 1),  # cranial right anteroposterior-direction
        (0, 2),  # posterior right cranial-caudal-direction
        (0, 3),  # right diagonal
        # (1, 2),  # This edge is not in this fiducial box
        (1, 3),  # anterior right cranial-caudal-direction
        (2, 3),  # caudal right anteroposterior-direction
        (4, 5),  # cranial left anteroposterior-direction
        (4, 6),  # posterior left cranial-caudal-direction
        (4, 7),  # left diagonal
        # (5, 6),  # This edge is not in the mr fiducial box
        (5, 7),  # anterior left cranial-caudal-direction
        (6, 7),  # caudal right anteroposterior-direction
    ]

    def get_edges(self, modality: str) -> list[tuple[int, int]]:
        if modality == "CT":
            return self.ct_edges
        elif modality == "MR":
            return self.mr_edges
        raise ModalityNotFoundError("This modality is not recognized")



import typer

import SimpleITK as sitk
from pathlib import Path
from typing import Optional
import logging

from stereotacticframe.frames import LeksellFrame
from stereotacticframe.frame_detector import FrameDetector
from stereotacticframe.slice_provider import AxialSliceProvider
from stereotacticframe.blob_detection import detect_blobs
from stereotacticframe.preprocessor import Preprocessor
from stereotacticframe.transforms import apply_transform

app = typer.Typer()

@app.command()
def calculate(
        input_image_path: Path,
        modality: str,
        output_transform_path: Optional[Path],
        log_dir: Optional[Path],
        logging_on: bool=False,
) -> None:
    if log_dir:
        fh = logging.FileHandler(log_dir)
        if logging_on:
            fh.setLevel(logging.DEBUG)
        logging.root.addHandler(fh)
    if logging_on:
        logging.root.setLevel(logging.DEBUG)

    # This could be generalized to any frame with a frame option    
    frame = LeksellFrame()

    preprocessor = Preprocessor(modality)

    provider = AxialSliceProvider(input_image_path, preprocessor)

    # bit anoying that I have to give modality as input for preprocessor and for framedetector
    detector = FrameDetector(frame, provider, detect_blobs, modality)

    detector.detect_frame()

    transform = detector.get_transform_to_frame_space()

    if not output_transform_path:
        output_transform_path = Path("./output.txt")
    
    sitk.WriteTransform(transform, output_transform_path)


@app.command()
def apply(image_path: Path, transform_path: Path, output_image_path: Path):
    frame = LeksellFrame()

    image = sitk.ReadImage(image_path)
    transform = sitk.ReadTransform(transform_path)

    sitk.WriteImage(apply_transform(image, transform, frame), output_image_path)


def main() -> None:
    logging.basicConfig(level=logging.INFO)
    app()


from typing import Protocol

class FrameProtocol(Protocol):
    dimensions: tuple[int, int, int]
    offset: tuple[float, float, float]
    direction: tuple[float, float, float, float, float, float, float, float, float]
    nodes: list[tuple[float, float, float]]
    ct_edges: list[tuple[int, int]]
    mr_edges: list[tuple[int, int]]

    def get_edges(self, modality: str) -> list[tuple[int, int]]:
        ...


import SimpleITK as sitk
from functools import partial, reduce
from typing import Callable

ImageToImageCallable = Callable[[sitk.Image], sitk.Image]


def _compose_two_functions(f: Callable, g: Callable):
    return lambda *a, **kw: f(g(*a, **kw))


def _compose(*fs):
    return reduce(_compose_two_functions, fs)

def _li_threshold() -> ImageToImageCallable:
    li = sitk.LiThresholdImageFilter()
    li.SetInsideValue(0)
    li.SetOutsideValue(1)
    li.SetNumberOfHistogramBins(256)
    return li.Execute

def _binary_dilate() -> ImageToImageCallable:
    dilate = sitk.BinaryDilateImageFilter()
    dilate.SetKernelRadius(2)
    dilate.SetKernelType(sitk.sitkBall)
    return dilate.Execute


_ct_pipeline = _compose(
    _binary_dilate(),
    partial(sitk.BinaryThreshold, lowerThreshold=1000, upperThreshold=10_000, insideValue=1, outsideValue=0),
)

_mr_pipeline = _compose(sitk.BinaryMorphologicalClosing, _li_threshold())

_threshold_map: dict[str, ImageToImageCallable] = {
    "CT": _ct_pipeline,
    "MR": _mr_pipeline,
}


class Preprocessor:
    def __init__(self, modality):
        self._modality: str = modality
        self._thresholder: ImageToImageCallable = _threshold_map[self._modality]

    def process(self, image: sitk.Image) -> sitk.Image:
        return self._thresholder(image)



from pathlib import Path
import SimpleITK as sitk
from typing import Protocol


def _reorient_rai(img):
    return sitk.DICOMOrient(img, "RAI")


class Processor(Protocol):
    def process(self, image: sitk.Image) -> sitk.Image:
        ...


class AxialSliceProvider:
    def __init__(self, image_path: Path, preprocessor: Processor):
        self._image_path: Path = image_path
        self._image: sitk.Image = preprocessor.process(sitk.ReadImage(self._image_path))
        self._rai_image: sitk.Image = _reorient_rai(self._image)
        self._counter: int = 0
        self._n_axial_slices: int = self._rai_image.GetSize()[-1]

    def next_slice(self):
        self._counter += 1
        return self._rai_image[..., self._counter - 1]

    def is_empty(self) -> bool:
        if self._counter >= self._n_axial_slices:
            return True
        return False

    def get_current_z_coordinate(self) -> float:
        point = self._rai_image.TransformIndexToPhysicalPoint([0, 0, self._counter - 1])
        return point[2]



import SimpleITK as sitk
import logging

def _get_label_statistics(label_img: sitk.Image, img: sitk.Image):
    label_statistics = sitk.LabelIntensityStatisticsImageFilter()
    label_statistics.Execute(label_img, img)
    return label_statistics


def detect_blobs(img_slice: sitk.Image) -> list[tuple[float, float]]:
    cc = sitk.ConnectedComponent(img_slice > 0)
    label_statistics = _get_label_statistics(cc, img_slice)
    blobs_list = []

    for label_idx in label_statistics.GetLabels():
        if not 0.5 < label_statistics.GetPhysicalSize(label_idx) < 30.:  # [mm²]
            continue
        
        blobs_list.append(label_statistics.GetCenterOfGravity(label_idx))

        logging.debug(f"Logging physical size: {label_statistics.GetPhysicalSize(label_idx)} for label {label_idx}")
        logging.debug(f"Logging elongation: {label_statistics.GetElongation(label_idx)} for label {label_idx}")
        logging.debug(f"Logging principal axes: {label_statistics.GetPrincipalAxes(label_idx)} for label {label_idx}")
        logging.debug(f"Logging principal moments: {label_statistics.GetPrincipalMoments(label_idx)} for label {label_idx}")
        logging.debug(f"Logging weighted elongation: {label_statistics.GetWeightedElongation(label_idx)} for label {label_idx}")
        logging.debug(f"Logging weighted principal axes: {label_statistics.GetWeightedPrincipalAxes(label_idx)} for label {label_idx}")
        logging.debug(f"Logging weighted principal moments: {label_statistics.GetWeightedPrincipalMoments(label_idx)} for label {label_idx}")

    return blobs_list



from __future__ import annotations

from typing import Protocol, Callable, Sequence
import SimpleITK as sitk
import numpy as np
import pyvista as pv
from vtk import vtkIterativeClosestPointTransform, vtkMatrix4x4
import logging

from stereotacticframe.frame_protocol import FrameProtocol

logger = logging.getLogger(__name__)


class SliceProviderProtocol(Protocol):
    def next_slice(self) -> sitk.Image:
        ...

    def is_empty(self) -> bool:
        ...

    def get_current_z_coordinate(self) -> float:
        ...


class PreprocessorProtocol(Protocol):
    def process(self, image: sitk.Image) -> sitk.Image:
        ...


BlobDetectorType = Callable[[sitk.Image], list[tuple[float, float]]]


def _create_lines(
        edges: list[tuple[int, int]], nodes: list[tuple[float, float, float]]
) -> pv.PolyData:
    line_mesh = pv.PolyData()
    for edge in edges:
        point0 = nodes[edge[0]]
        point1 = nodes[edge[1]]
        line_mesh += pv.Line(point0, point1)  # type: ignore
    return line_mesh  # type: ignore


def _iterative_closest_point(
        source: pv.PolyData,
        target: pv.PolyData,
        iterations: int,
        number_of_landmarks: int = 20_000,
        start_by_mathing_centroids: bool = True,
) -> vtkIterativeClosestPointTransform:
    icp = vtkIterativeClosestPointTransform()
    icp.SetSource(source)
    icp.SetTarget(target)
    icp.SetMaximumNumberOfIterations(iterations)
    icp.SetMaximumNumberOfLandmarks(number_of_landmarks)
    icp.SetMeanDistanceModeToAbsoluteValue()
    icp.StartByMatchingCentroidsOff()
    if start_by_mathing_centroids:
        icp.StartByMatchingCentroidsOn()
    icp.GetLandmarkTransform().SetModeToRigidBody()
    icp.Update()
    return icp

def _transform4x4_to_sitk_affine(vtk_matrix: vtkMatrix4x4) -> sitk.Transform:
    dimension = 3  # dimension is always 3 in a 4x4 transform
    affine = sitk.AffineTransform(dimension)
    parameters = list(affine.GetParameters())
    for i in range(dimension):
        for j in range(dimension):
            parameters[i * dimension + j] = vtk_matrix.GetElement(i, j)
    for i in range(3):
        parameters[i + dimension * dimension] = vtk_matrix.GetElement(
            i, dimension
        )
    affine.SetParameters(parameters)
    return affine


def calculate_frame_extent_3d(
        frame_dimensions: tuple[float, float, float],
        voxel_spacing: tuple[float, float, float],
        offset: tuple[float, float, float],
) -> tuple[int, ...]:
    extent = tuple()
    for dim in range(len(frame_dimensions)):
        frame_dim = frame_dimensions[dim]
        current_voxel_spacing = voxel_spacing[dim]
        offs = offset[dim]
        extent += (round((frame_dim + 2 * abs(offs)) / current_voxel_spacing),)

    return extent


class FrameDetector:
    def __init__(
            self,
            frame: FrameProtocol,
            slice_provider: SliceProviderProtocol,
            blob_detector: BlobDetectorType,
            modality: str,
    ):
        self._frame = frame
        self._slice_provider = slice_provider
        self._blob_detector = blob_detector
        self._point_cloud: pv.PolyData | None = None
        self._sitk_transform: sitk.Transform | None = None
        self._frame_object: pv.PolyData = _create_lines(frame.get_edges(modality), frame.nodes)
        self._modality = modality

    # Quite a bit of cohesion here, not sure if it's a problem, since it has to come together somewhere
    def detect_frame(self) -> None:
        blobs_list = []
        while not self._slice_provider.is_empty():
            next_slice = self._slice_provider.next_slice()
            blobs_list += [
                two_d_point + (self._slice_provider.get_current_z_coordinate(),)
                for two_d_point in self._blob_detector(next_slice)
            ]
        self._point_cloud = pv.PolyData(np.asarray(blobs_list))

    def get_transform_to_frame_space(self) -> sitk.Transform:

        if self._point_cloud is None:
            raise ValueError("Detect frame was not run or there is a problem with detect frame.")
       
       
        point_cloud = self._point_cloud.copy()
        # Run 1 iteration to do centroid allignment 
        centroid_transform = _iterative_closest_point(point_cloud, self._frame_object, iterations=1)
        
        centroid_matrix = centroid_transform.GetMatrix()
        centroid_inverse_matrix = centroid_matrix.NewInstance()
        centroid_inverse_matrix.DeepCopy(centroid_matrix)
        centroid_inverse_matrix.Invert()
        
        point_cloud.transform(centroid_matrix)
        
        # Very liberally clean some points 
        new_points = point_cloud.points
        
        if self._modality == "CT":
            # CT typically has some noise from the CT fiducial box clamp
            # This messes with the initial allignment being too high
            new_points = new_points[new_points[..., 2] > -100]  # Remove high points
            new_points = new_points[new_points[..., 2] < 20]  # Remove low points
        
        else:
            new_points = new_points[new_points[..., 2] > -120]  # Remove high points
            new_points = new_points[new_points[..., 2] < 0]  # Remove low points

        right_points = new_points[new_points[..., 0] < 30]  # Keep only right points
        left_points = new_points[new_points[..., 0] > 160]  # Keep only left points
        
        new_cloud = pv.PolyData(right_points) + pv.PolyData(left_points)
        
        new_cloud.transform(centroid_inverse_matrix)
        
        initial_transform = _iterative_closest_point(new_cloud, self._frame_object, 20000)
        point_cloud = self._point_cloud.copy()

        initial_matrix = initial_transform.GetMatrix()
        initial_inverse_matrix = initial_matrix.NewInstance()
        initial_inverse_matrix.DeepCopy(initial_matrix)
        initial_inverse_matrix.Invert()
        new_cloud.transform(initial_inverse_matrix)
        
        point_cloud.transform(initial_matrix)
       
        # Remove upper 20mm and lower 20mm
        new_points = point_cloud.points
        new_points = new_points[new_points[..., 2] > -110]  # Remove top 10 mm
        new_points = new_points[new_points[..., 2] < -10]  # Remove bottom 10 mm

        right_points = new_points[new_points[..., 0] < 10]  # Keep only right points
        left_points = new_points[new_points[..., 0] > 180]  # Keep only left points
        new_cloud = pv.PolyData(right_points) + pv.PolyData(left_points)
        
        new_cloud.transform(initial_inverse_matrix)

        refined_transform = _iterative_closest_point(new_cloud, self._frame_object, iterations=10_000)
        refined_matrix = refined_transform.GetMatrix()
        refined_inverse_matrix = refined_matrix.NewInstance()
        refined_inverse_matrix.DeepCopy(refined_matrix)
        refined_inverse_matrix.Invert()
        
        point_cloud = self._point_cloud.copy()
        point_cloud.transform(refined_matrix)
        
        new_cloud.transform(refined_matrix)
        
        closest_points = self._calculate_closest_points_in_frame_to(point_cloud.points.copy())

        exact_distance = np.linalg.norm(point_cloud.points - closest_points, axis=1)
        
        new_points = point_cloud.points
        new_points = new_points[np.nonzero(exact_distance < 3.0)]
        new_points = new_points[new_points[..., 2] > -110]  # Remove top 10 mm
        new_points = new_points[new_points[..., 2] < -10]  # Remove bottom 10 mm
        right_points = new_points[new_points[..., 0] < 10]  # Keep only right points
        left_points = new_points[new_points[..., 0] > 180]  # Keep only left points
        new_cloud = pv.PolyData(right_points) + pv.PolyData(left_points)

        new_cloud.transform(refined_inverse_matrix)
        
        final_transform = _iterative_closest_point(new_cloud, self._frame_object, iterations=10_000)
         
        final_matrix = final_transform.GetMatrix()
        final_inverse_matrix = final_matrix.NewInstance()
        final_inverse_matrix.DeepCopy(final_matrix)
        final_inverse_matrix.Invert()
        new_cloud.transform(final_matrix)
        
        point_cloud = self._point_cloud.copy()
        point_cloud.transform(final_matrix)

        closest_points_in_frame = self._calculate_closest_points_in_frame_to(new_cloud.points)
        self._set_mean_max(closest_points_in_frame, new_cloud.points)
        
        logger.info(f"Mean detection error: {self._mean_error}") 
        logger.info(f"Max detection error: {self._max_error}") 

        
        final_itk_transform = _transform4x4_to_sitk_affine(final_matrix)
        return final_itk_transform.GetInverse()
    
    def _calculate_closest_points_in_frame_to(self, points: pv.NumpyArray) -> pv.NumpyArray:
        _, closest_points = self._frame_object.find_closest_cell(
            points, return_closest_point=True
        )
        return closest_points
    
    def _set_mean_max(self, points: pv.NumpyArray, poly_points: pv.NumpyArray) -> None:
        distances = np.linalg.norm(points - poly_points, axis=1)
        self._mean_error = distances.mean()
        self._max_error = distances.max() 